import pathlib
import platform
import tarfile
import tempfile
from typing import override

import httpx

from terraflex.utils.dependency_downloader import DownloaderProtocol, mv_executable_to_dest

# pylint: disable=line-too-long
AGE_URL_DOWNLOAD = "https://github.com/FiloSottile/age/releases/download/v{version}/age-v{version}-{platform}.tar.gz"


class AgeDownloader(DownloaderProtocol):
    def _get_platform_name(self):
        current_platform = platform.system().lower()
        machine_type = platform.machine()
        # check if amd64 or x86_64
        if machine_type in ["AMD64", "x86_64"]:
            return f"{current_platform}-amd64"

        return f"{current_platform}-arm64"

    @override
    async def __call__(self, version: str, expected_paths: dict[str, pathlib.Path]) -> None:
        url = AGE_URL_DOWNLOAD.format(
            version=version,
            platform=self._get_platform_name(),
        )
        print("downloading age from", url)
        async with httpx.AsyncClient() as client:
            response = await client.get(
                url,
                follow_redirects=True,
            )

        if response.status_code != 200:
            raise RuntimeError(f"Failed to download age: {response.status_code} - {response.text}")

        with tempfile.TemporaryDirectory() as tmp_dir:
            with tempfile.NamedTemporaryFile() as temp:
                temp.write(response.content)
                temp.flush()

                with tarfile.open(temp.name, "r") as tar:
                    tar.extractall(tmp_dir)

            age_bin = pathlib.Path(tmp_dir) / "age" / "age"
            if "age" not in expected_paths:
                raise ValueError("Expected age to be in the expected paths")

            mv_executable_to_dest(age_bin, expected_paths["age"])

            age_keygen = pathlib.Path(tmp_dir) / "age" / "age-keygen"
            if "age-keygen" not in expected_paths:
                raise ValueError("Expected age-keygen to be in the expected paths")

            mv_executable_to_dest(age_keygen, expected_paths["age-keygen"])



import questionary

from terraflex.plugins.git_storage_provider.git_storage_provider import (
    GitStorageProviderInitConfig,
    GitStorageProviderItemIdentifier,
)
from terraflex.server.config import (
    StorageProviderConfig,
    StorageProviderUsageConfig,
)


async def build_git_storage_provider() -> StorageProviderConfig:
    origin_url = await questionary.path(
        "What is the origin url?",
    ).ask_async()

    return StorageProviderConfig(
        type="git",
        **GitStorageProviderInitConfig(origin_url=origin_url).model_dump(),
    )


async def build_git_key_identifier(
    provider_name: str, path: str, question: str = "What is the location of the file?"
) -> StorageProviderUsageConfig:
    path = await questionary.text(
        question,
        default=path,
    ).ask_async()

    return StorageProviderUsageConfig(
        provider=provider_name,
        params=GitStorageProviderItemIdentifier(path=path).model_dump(),
    )



import asyncio
import multiprocessing
import pathlib
import subprocess
import time
from contextlib import contextmanager
from typing import Annotated, Any, Iterator

import httpx
import questionary
import typer
import yaml
from uvicorn import Config, Server

from terraflex.cli.builders.wizard import start_configfile_creation_wizard
from terraflex.server.app import (
    CONFIG_FILE_NAME,
    create_storage_providers,
    initialize_manager,
    start_server,
)
from terraflex.server.app import (
    app as server_app,
)
from terraflex.server.app import (
    config as server_config,
)
from terraflex.server.config import ConfigFile
from terraflex.server.storage_provider_base import LockableStorageProviderProtocol

READY_MESSAGE = """\
backend "http" {{
{content}
}}
"""

ADDRESS_INFO = """\
    address = "http://localhost:{port}/{stack_name}/state"
"""

LOCK_INFO = """\
    lock_address = "http://localhost:{port}/{stack_name}/lock"
    lock_method = "PUT"
    unlock_address = "http://localhost:{port}/{stack_name}/lock"
    unlock_method = "DELETE"
"""


app = typer.Typer(pretty_exceptions_enable=False, rich_markup_mode=None)


@contextmanager
def capture_aborts() -> Iterator[None]:
    try:
        yield
    except typer.Abort as e:
        print("Error:", e)
        raise


async def _init() -> None:
    port = 8600
    manager = await initialize_manager()
    config_file_location = pathlib.Path(CONFIG_FILE_NAME)
    if config_file_location.exists():
        print("Configuration file already exists")
        should_replace = await questionary.confirm(
            "Do you want to replace it?",
            default=False,
        ).ask_async()
        if not should_replace:
            print("Aborting...")
            return

        print("Replacing existing configuration file")

    stack_name, result_file = await start_configfile_creation_wizard(manager)
    raw_file = yaml.safe_dump(yaml.safe_load(result_file.model_dump_json()))
    config_file_location.write_text(raw_file, encoding="utf-8")

    print("\n\n")
    print("Configuration file created")
    print("You can now start the server with `terraflex start`")

    await print_binding_message(stack_name, port)


@app.command()
def init() -> None:
    """Initialize the configuration file for the server in current directory.

    Starts an interactive wizard to create the configuration file.

    Output will be a file named `terraflex.yaml` in the current directory.
    """
    with capture_aborts():
        asyncio.run(_init())


@app.command()
def start(
    port: Annotated[int, typer.Option(help="Port to run the server on")] = 8600,
) -> None:
    """Starts the server with the configuration file in the current directory."""
    start_server(port)


async def print_binding_message(stack_name: str, port: int) -> None:
    manager = await initialize_manager()
    config_file = pathlib.Path(CONFIG_FILE_NAME)
    if not config_file.exists():
        raise FileNotFoundError(f"Config file not found: {config_file}")

    content = config_file.read_text(encoding="utf-8")
    config_dict = yaml.safe_load(content)
    config = ConfigFile.model_validate(config_dict)

    if stack_name not in config.stacks:
        raise ValueError(f"Stack not found: {stack_name}")

    storage_provider_name = config.stacks[stack_name].state_storage.provider
    storage_providers = await create_storage_providers(config, manager=manager, workdir=server_config.state_dir)
    content_parts = [
        ADDRESS_INFO.format(port=port, stack_name=stack_name).rstrip(),
    ]

    if isinstance(storage_providers[storage_provider_name], LockableStorageProviderProtocol):
        content_parts.append(LOCK_INFO.format(port=port, stack_name=stack_name).rstrip())

    print("In terraform backend configuration, use the following:\n")
    print(READY_MESSAGE.format(content="\n".join(content_parts)))


@app.command()
def print_bindings(
    stack_name: Annotated[str, typer.Argument(help="Name of the stack")],
    port: Annotated[int, typer.Option(help="Port to run the server on")] = 8600,
) -> None:
    """Prints the terraform backend configuration for the given stack name and port."""
    with capture_aborts():
        asyncio.run(print_binding_message(stack_name, port))


class UvicornServer(multiprocessing.Process):
    def __init__(self, config: Config):
        super().__init__()
        self.server = Server(config=config)
        self.config = config

    def stop(self):
        self.terminate()

    def run(self, *args: list[Any], **kwargs: dict[str, Any]):
        self.server.run()


def wait_until_ready(port: int):
    client = httpx.Client()
    while True:
        try:
            response = client.get(f"http://localhost:{port}/ready")
            response.raise_for_status()
            break
        except Exception:
            time.sleep(0.2)


@app.command()
def wrap(
    args: Annotated[list[str], typer.Argument(help="Command to run")],
    verbose: Annotated[
        bool,
        typer.Option("-v", "--verbose", help="Print more details about the backend"),
    ] = False,
    port: Annotated[int, typer.Option(help="Port to run the server on")] = 8600,
) -> None:
    """Main command that allows wrapping any command with the context of the server running.

    Its main purpose is to allow running terraform commands with the server running.

    Examples:
    $ terraflex wrap -- terraform init
    """
    instance = UvicornServer(
        config=Config(
            app=server_app,
            port=port,
            access_log=verbose,
            log_level="info" if verbose else "warning",
        )
    )
    instance.start()
    wait_until_ready(port)
    # run the command
    subprocess.run(args)
    instance.stop()


def main() -> None:
    app()



import asyncio
import pathlib
from typing import Collection, Mapping, Optional


class BinaryController:
    def __init__(
        self,
        binary_location: pathlib.Path,
        cwd: Optional[pathlib.Path] = None,
        env: Optional[Mapping[str, str]] = None,
    ):
        self.binary_location = binary_location
        self.cwd = cwd
        self.env = env or {}

    async def _execute_command(
        self,
        args: Collection[str | bytes],
        stdin: Optional[bytes] = None,
    ) -> bytes:
        proc = await asyncio.create_subprocess_exec(
            self.binary_location,
            *args,
            cwd=self.cwd,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=self.env,
        )

        stdout, stderr = await proc.communicate(stdin)
        if proc.returncode != 0:
            raise RuntimeError(f"Failed to execute binary: {stderr}")

        return stdout



import abc
import pathlib
from contextlib import contextmanager
from typing import Any, Iterator, Protocol, Self, TypeVar, runtime_checkable

from pydantic import BaseModel
from terraflex.server.base_state_lock_provider import LockBody, LockingError
from terraflex.utils.dependency_manager import DependenciesManager

STORATE_PROVIDERS_ENTRYPOINT = "terraflex.plugins.storage_provider"


class ItemKey(BaseModel, abc.ABC):
    """Params required to reference an item in a storage provider.

    Every storage provider must implement a subclass of `ItemKey` to represent the key of the item - using the validate_key() method.
    """

    @abc.abstractmethod
    def as_string(self) -> str:
        """Return the string representation of the key."""
        ...


T = TypeVar("T", bound=BaseModel)


def parse_item_key(item_key: ItemKey, model: type[T]) -> T:
    if not isinstance(item_key, model):
        raise ValueError(f"Item key is not of type {model}")

    return item_key


@runtime_checkable
class StorageProviderProtocol(Protocol):
    """Protocol for storage providers - Read only.

    Readable storage is the most basic storage provider - it allows to read files from the storage.

    Every readonly storage provider must implement `StorageProviderProtocol` methods -
    and register to the `terraflex.plugins.storage_provider` entrypoint.
    """

    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        manager: DependenciesManager,
        workdir: pathlib.Path,
    ) -> Self:
        """Create an instance of the storage provider from the configuration.

        Args:
            raw_config: The raw configuration propagated from the storage provider config.
            manager: The dependencies manager - allows to request a binary path from.
            workdir: The data directory of terraflex - located at `~/.local/share/terraflex` -
                can be used to manage state of the provider
        """
        ...

    @classmethod
    def validate_key(cls, key: dict[str, Any]) -> ItemKey:
        """Validate the key of the item in the storage provider from a config of usage.

        Args:
            key: a dict with parameters to build the key of the item.

        Returns:
            The validated key.
        """
        ...

    # read related
    async def get_file(self, item_identifier: ItemKey) -> bytes:
        """Get the content of the file.

        Args:
            item_identifier: The identifier of the file.
        """
        ...


@runtime_checkable
class WriteableStorageProviderProtocol(StorageProviderProtocol, Protocol):
    """Protocol for storage providers - Writeable.

    Writeable storage provider allows to write files to the storage.

    Every writeable storage provider must implement `WriteableStorageProviderProtocol` methods and its parent methods -
        and register to the `terraflex.plugins.storage_provider` entrypoint.
    """

    # write related
    async def put_file(self, item_identifier: ItemKey, data: bytes) -> None:
        """Put the content of the file in the provided file.

        Args:
            item_identifier: The identifier of the file.
            data: The content of the file.
        """
        ...

    async def delete_file(self, item_identifier: ItemKey) -> None:
        """Delete the file.

        Args:
            item_identifier: The identifier of the file.
        """
        ...


@runtime_checkable
class LockableStorageProviderProtocol(WriteableStorageProviderProtocol, Protocol):
    """Protocol for storage providers - Lockable.

    Lockable storage provider allows to lock items to prevent concurrent writes.
    Allows to support terraform state locking - see more at official [docs](https://developer.hashicorp.com/terraform/language/state/locking).

    Every lockable storage provider must implement `LockableStorageProviderProtocol` methods and its parent methods -
        and register to the `terraflex.plugins.storage_provider` entrypoint.
    """

    # lock related
    async def read_lock(self, item_identifier: ItemKey) -> LockBody:
        """Read the lock of the item.

        Args:
            item_identifier: The identifier of the item.

        Returns:
            The lock data of the item.
        """
        ...

    async def acquire_lock(self, item_identifier: ItemKey, data: LockBody) -> None:
        """Acquire the lock of the item.

        Args:
            item_identifier: The identifier of the item.
            data: The lock data of the item.
        """
        ...

    async def release_lock(self, item_identifier: ItemKey) -> None: ...


@contextmanager
def assume_lock_conflict_on_error(lock_id: str) -> Iterator[None]:
    try:
        yield
    except Exception as e:
        raise LockingError(
            "Failed to lock state",
            lock_id=lock_id,
        ) from e



import tempfile
from pathlib import Path

from terraflex.utils.binary_controller import BinaryController


class AgeController(BinaryController):
    def __init__(
        self,
        binary_location: Path,
        private_key: bytes,
        public_key: bytes,
    ):
        super().__init__(binary_location)
        self.private_key = private_key
        self.public_key = public_key

    async def encrypt(self, _: str, content: bytes) -> bytes:
        return await self._execute_command(
            [
                "--encrypt",
                "-r",
                self.public_key,
            ],
            stdin=content,
        )

    async def decrypt(self, _: str, content: bytes) -> bytes:
        with tempfile.NamedTemporaryFile() as temp:
            temp.write(self.private_key)
            temp.flush()

            return await self._execute_command(
                [
                    "--decrypt",
                    "-i",
                    temp.name,
                ],
                stdin=content,
            )


class AgeKeygenController(BinaryController):
    async def generate_key_bytes(self) -> bytes:
        return (await self._execute_command([])).strip()

    async def generate_key(self, key_location: Path) -> None:
        key_bytes = await self.generate_key_bytes()
        key_location.write_bytes(key_bytes)

    async def get_public_key(self, key_location: Path) -> bytes:
        key = key_location.read_bytes()
        return await self.get_public_key_from_bytes(key)

    async def get_public_key_from_bytes(self, content: bytes) -> bytes:
        return (await self._execute_command(["-y"], stdin=content)).strip()



import json
from dataclasses import dataclass

from terraflex.server.base_state_lock_provider import (
    Data,
    LockBody,
    LockingError,
    StateLockProviderProtocol,
)
from terraflex.server.storage_provider_base import (
    ItemKey,
    LockableStorageProviderProtocol,
    WriteableStorageProviderProtocol,
)
from terraflex.server.transformation_base import (
    TransformerProtocol,
)


@dataclass
class TFStack:
    name: str
    data_transformers: list[TransformerProtocol]
    storage_driver: WriteableStorageProviderProtocol
    state_file_storage_identifier: ItemKey


class TFStateLockController(StateLockProviderProtocol):
    def __init__(
        self,
        stacks: dict[str, TFStack],
    ):
        self.stacks = stacks

    def _validate_stack(self, stack_name: str) -> TFStack:
        stack = self.stacks.get(stack_name)
        if stack is None:
            raise ValueError(f"Undeclared stack: {stack_name}")

        return stack

    async def get(self, stack_name: str) -> Data | None:
        stack = self._validate_stack(stack_name)
        try:
            data = await stack.storage_driver.get_file(stack.state_file_storage_identifier)

        except FileNotFoundError:
            return None

        content = data
        for transformer in reversed(stack.data_transformers):
            content = await transformer.transform_read_file_content(
                stack.state_file_storage_identifier.as_string(), content
            )

        return json.loads(content)

    async def put(self, stack_name: str, lock_id: str, value: Data) -> None:
        stack = self._validate_stack(stack_name)
        await self._check_lock(stack_name, lock_id)
        # lock is locked by me

        data = json.dumps(value).encode()
        for transformer in stack.data_transformers:
            data = await transformer.transform_write_file_content(stack.state_file_storage_identifier.as_string(), data)

        await stack.storage_driver.put_file(stack.state_file_storage_identifier, data)

    async def delete(self, stack_name: str, lock_id: str) -> None:
        stack = self._validate_stack(stack_name)
        await self._check_lock(stack_name, lock_id)
        # lock is locked by me

        await stack.storage_driver.delete_file(stack.state_file_storage_identifier)

    async def read_lock(self, stack_name: str) -> LockBody | None:
        stack = self._validate_stack(stack_name)
        if not isinstance(stack.storage_driver, LockableStorageProviderProtocol):
            raise NotImplementedError("This storage provider does not support writing")

        try:
            return await stack.storage_driver.read_lock(stack.state_file_storage_identifier)

        except FileNotFoundError:
            return None

    async def _check_lock(self, stack_name: str, lock_id: str) -> LockBody:
        stack = self._validate_stack(stack_name)
        if not isinstance(stack.storage_driver, LockableStorageProviderProtocol):
            # This storage provider does not support locking
            return LockBody(
                ID="0000000000000000000", Operation="read", Who="me", Version="1", Created="2000-01-01T00:00:00Z"
            )

        data = await self.read_lock(stack_name)
        if data is None:
            raise LockingError(
                "Failed to lock state - no lock is present",
                lock_id=lock_id,
            )

        if data.ID != lock_id:
            raise LockingError(
                "Failed to lock state - someone else has already locked it",
                lock_id=lock_id,
            )

        return data

    async def lock(self, stack_name: str, data: LockBody) -> None:
        stack = self._validate_stack(stack_name)
        if not isinstance(stack.storage_driver, LockableStorageProviderProtocol):
            return

        await stack.storage_driver.acquire_lock(stack.state_file_storage_identifier, data)

    async def unlock(self, stack_name: str) -> None:
        stack = self._validate_stack(stack_name)
        if not isinstance(stack.storage_driver, LockableStorageProviderProtocol):
            return

        await stack.storage_driver.release_lock(stack.state_file_storage_identifier)



from typing import Any, Protocol, Self, runtime_checkable

from terraflex.server.storage_provider_base import StorageProviderProtocol
from terraflex.utils.dependency_manager import DependenciesManager


@runtime_checkable
class EncryptionProtocol(Protocol):
    """Protocol for encryption providers.


    Every encryption provider must implement `EncryptionProtocol` methods - and register to the `terraflex.plugins.transformer.encryption` entrypoint.

    Example:
        Register age encryption provider - if your project is based on poetry:
        ```toml
        [tool.poetry.plugins."terraflex.plugins.transformer.encryption"]
        age = "terraflex.plugins.encryption_transformation.age.provider:AgeEncryptionProvider"
        ```
    """

    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        storage_providers: dict[str, StorageProviderProtocol],
        manager: DependenciesManager,
    ) -> Self:
        """Create an instance of the encryption provider from the configuration.

        Args:
            raw_config: The raw configuration propagated from the transformer config.
            storage_providers: All the initialized storage providers specified in the config file.
            manager: The dependencies manager - allows to request a binary path from.

        Returns:
            The initialized instance of the encryption provider.
        """
        ...

    async def encrypt(self, file_name: str, content: bytes) -> bytes:
        """Encrypt the content of the file.

        Args:
            file_name: The name of the file.
            content: The content of the file.

        Returns:
            The encrypted content.
        """
        ...

    async def decrypt(self, file_name: str, content: bytes) -> bytes:
        """Decrypt the content of the file.

        Args:
            file_name: The name of the file.
            content: The content of the file.

        Returns:
            The decrypted content.
        """
        ...



import pathlib
from typing import Any, Protocol, Self, runtime_checkable

from terraflex.server.storage_provider_base import StorageProviderProtocol
from terraflex.utils.dependency_manager import DependenciesManager

TRANSFORMERS_ENTRYPOINT = "terraflex.plugins.transformer"


@runtime_checkable
class TransformerProtocol(Protocol):
    """Protocol for state transformation providers.

    Every state transformation provider must implement `TransformerProtocol` methods - and register to the `terraflex.plugins.transformer` entrypoint.

    Example:
        Register encryption transformer - if your project is based on poetry:
        ```toml
        [tool.poetry.plugins."terraflex.plugins.transformer"]
        encryption = "terraflex.plugins.encryption_transformation.encryption_transformation_provider:EncryptionTransformation"
        ```
    """

    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        storage_providers: dict[str, StorageProviderProtocol],
        manager: DependenciesManager,
        workdir: pathlib.Path,
    ) -> Self:
        """Create an instance of the transformation provider from the configuration.

        Args:
            raw_config: The raw configuration propagated from the transformer config.
            storage_providers: All the initialized storage providers specified in the config file.
            manager: The dependencies manager - allows to request a binary path from.
            workdir: The data directory of terraflex - located at `~/.local/share/terraflex` -
                can be used to manage state of the provider.
        """
        ...

    async def transform_write_file_content(self, file_identifier: str, content: bytes) -> bytes:
        """Transform the content of the file before writing it to the storage provider.

        Args:
            file_identifier: The identifier of the file - calculated by calling to_string()
                method of the storage usage params.
            content: The content of the file.
        """
        ...

    async def transform_read_file_content(self, file_identifier: str, content: bytes) -> bytes:
        """Transform the content of the file after reading it from the storage provider.

        Args:
            file_identifier: The identifier of the file - calculated by calling to_string()
                method of the storage usage params.
            content: The content of the file.
        """
        ...



import pathlib
import subprocess
from contextlib import suppress
from typing import Any, Optional, Self, override

from pydantic import BaseModel
from terraflex.server.base_state_lock_provider import LockBody
from terraflex.server.storage_provider_base import (
    ItemKey,
    LockableStorageProviderProtocol,
    assume_lock_conflict_on_error,
    parse_item_key,
)
from terraflex.utils.dependency_manager import DependenciesManager


class GitStorageProviderItemIdentifier(ItemKey):
    """Params required to reference an item in Git storage provider.

    Attributes:
        path: The path to a specific file relative to the repository root,
            folders are also allowed as part of the path.
    """

    path: str

    @override
    def as_string(self) -> str:
        return self.path


class GitStorageProviderInitConfig(BaseModel):
    """Initialization params required to initialize Git storage provider.

    Attributes:
        origin_url: The URL of the git repository. Must be accessible by the current user.
            Example: git@github.com:IamShobe/tf-state.git
        ref: The branch to use.
        clone_path: The path to clone the repository to. Default: None. (will be set to ~/.local/share/terraflex/git_storage/<repo_name>)
    """

    origin_url: str
    ref: str = "main"
    clone_path: Optional[pathlib.Path] = None


def directory_is_empty(directory: pathlib.Path) -> bool:
    return not any(directory.iterdir())


class GitStorageProvider(LockableStorageProviderProtocol):
    """This follows the steps described in the suggestion here:
    https://github.com/plumber-cd/terraform-backend-git
    """

    def __init__(
        self,
        origin_url: str,
        clone_path: pathlib.Path,
        ref: str = "main",
    ) -> None:
        self.clone_path = clone_path.expanduser()
        self.origin_url = origin_url
        self.ref = ref

        self._initialize()

    def _initialize(self) -> None:
        # create parent directories
        self.clone_path.parent.mkdir(parents=True, exist_ok=True)

        if not self.clone_path.exists():
            # clone the repository
            self._git("clone", self.origin_url, str(self.clone_path), cwd=self.clone_path.parent)

        self.validate()

    # TODO: Add invoking for this method...
    def validate(self) -> None:
        # check that the path isn't dirty
        if not self.clone_path.exists():
            raise FileNotFoundError(f"Path {self.clone_path} does not exist")

        if not self.clone_path.is_dir():
            raise NotADirectoryError(f"Path {self.clone_path} is not a directory")

        if not (self.clone_path / ".git").exists():
            raise FileNotFoundError(f"Path {self.clone_path} is not a git repository")

        if self._is_dirty():
            raise RuntimeError(
                f"Path {self.clone_path} is dirty - please commit or stash changes before using this provider"
            )

    @override
    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        manager: DependenciesManager,
        workdir: pathlib.Path,
    ) -> Self:
        result = GitStorageProviderInitConfig.model_validate(raw_config)

        repo_name = result.origin_url.split("/")[-1].replace(".git", "")
        result.clone_path = result.clone_path or (workdir / "git_storage" / repo_name)

        return cls(
            **result.model_dump(),
        )

    def _git(self, command: str, *args: str, cwd: Optional[str | pathlib.Path] = None) -> str:
        proc = subprocess.run(
            ["git", command, *args],
            cwd=cwd or self.clone_path,
            capture_output=True,
            text=True,
            check=False,
        )

        if proc.returncode != 0:
            raise RuntimeError(f"Error running git command: {proc.stderr}")

        return proc.stdout

    def _cleanup_workspace(self) -> None:
        self._git("reset", "--hard")
        self._git("checkout", self.ref)

    def _is_dirty(self) -> bool:
        return bool(self._git("status", "--porcelain"))

    @override
    @classmethod
    def validate_key(cls, key: dict[str, Any]) -> GitStorageProviderItemIdentifier:
        return GitStorageProviderItemIdentifier.model_validate(key)

    @override
    async def get_file(self, item_identifier: ItemKey) -> bytes:
        parsed_key = parse_item_key(item_identifier, GitStorageProviderItemIdentifier)
        file_name = parsed_key.path
        self._cleanup_workspace()
        # pull latest changes
        self._git("pull", "origin", self.ref)

        # read state
        state_file = self.clone_path / file_name
        try:
            return state_file.read_bytes()

        except FileNotFoundError as exc:
            raise FileNotFoundError(f"File {file_name} not found in the repository") from exc

    def commit_and_push_changes(self, message: str, ref: Optional[str] = None) -> None:
        self._git("add", ".")
        self._git("commit", "-m", message)
        self._git("push", "origin", ref or self.ref)

    @override
    async def put_file(self, item_identifier: ItemKey, data: bytes) -> None:
        parsed_key = parse_item_key(item_identifier, GitStorageProviderItemIdentifier)
        file_name = parsed_key.path
        self._cleanup_workspace()
        # pull latest changes
        self._git("pull", "origin", self.ref)

        # save state
        state_file = self.clone_path / file_name
        state_file.parent.mkdir(parents=True, exist_ok=True)
        state_file.write_bytes(data)

        self.commit_and_push_changes(f"Update state - {file_name}")

    @override
    async def delete_file(self, item_identifier: ItemKey) -> None:
        parsed_key = parse_item_key(item_identifier, GitStorageProviderItemIdentifier)
        file_name = parsed_key.path
        self._cleanup_workspace()
        # pull latest changes
        self._git("pull", "origin", self.ref)

        # delete state
        state_file = self.clone_path / file_name
        state_file.unlink()

        self.commit_and_push_changes(f"Delete state - {file_name}")

    @override
    async def read_lock(self, item_identifier: ItemKey) -> LockBody:
        parsed_key = parse_item_key(item_identifier, GitStorageProviderItemIdentifier)
        file_name = parsed_key.path
        self._cleanup_workspace()
        # delete lock branch if it exists
        with suppress(Exception):
            self._git("branch", "-D", f"locks/{file_name}")
        # pull latest changes
        self._git("fetch", "origin", "refs/heads/locks/*:refs/remotes/origin/locks/*")

        try:
            self._git("checkout", f"locks/{file_name}")

        except RuntimeError as exc:
            raise FileNotFoundError(f"Lock file {file_name} not found in the repository") from exc

        # read lock data
        lock_file = self.clone_path / "locks" / "lock.lock"
        return LockBody.model_validate_json(lock_file.read_bytes())

    @override
    async def acquire_lock(self, item_identifier: ItemKey, data: LockBody) -> None:
        parsed_key = parse_item_key(item_identifier, GitStorageProviderItemIdentifier)
        file_name = parsed_key.path
        self._cleanup_workspace()
        # delete lock branch if it exists
        with suppress(Exception):
            self._git("branch", "-D", f"locks/{file_name}")
        # pull latest changes
        self._git("pull", "origin", self.ref)

        # create a new locking branch
        self._git("checkout", "-b", f"locks/{file_name}")

        # make sure lock folder exists
        locks_dir = self.clone_path / "locks"
        locks_dir.mkdir(exist_ok=True)
        # write lock file
        lock_file = locks_dir / "lock.lock"
        lock_file.write_bytes(data.model_dump_json().encode())

        self._git("add", str(lock_file))
        self._git("commit", "-m", f"Locking state - id {data.ID}")
        with assume_lock_conflict_on_error(lock_id=data.ID):
            self._git("push", "origin", f"locks/{file_name}")

    @override
    async def release_lock(self, item_identifier: ItemKey) -> None:
        parsed_key = parse_item_key(item_identifier, GitStorageProviderItemIdentifier)
        file_name = parsed_key.path
        self._git("push", "origin", "--delete", f"locks/{file_name}")



import pathlib
import subprocess
from typing import Any, Self, override

from pydantic import BaseModel
from terraflex.server.storage_provider_base import (
    ItemKey,
    StorageProviderProtocol,
    parse_item_key,
)
from terraflex.utils.dependency_manager import DependenciesManager


class OnePasswordProviderItemIdentifier(ItemKey):
    """Params required to reference an item in 1Password storage provider.

    Attributes:
        reference_uri: 1Password URI to the item. Example: `op://<vault>/<item>/<field>`
    """

    reference_uri: str

    @override
    def as_string(self) -> str:
        return self.reference_uri.replace("/", "_").replace(":", "_")


class OnePasswordStorageProviderInitConfig(BaseModel):
    """Initialization params required to initialize 1Password storage provider.

    1Password storage provider currently have no initialization params required.
    """


class OnePasswordStorageProvider(StorageProviderProtocol):
    @override
    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        manager: DependenciesManager,
        workdir: pathlib.Path,
    ) -> Self:
        result = OnePasswordStorageProviderInitConfig.model_validate(raw_config)
        cls._validate_binary()
        return cls(
            **result.model_dump(),
        )

    @classmethod
    def _validate_binary(cls):
        try:
            subprocess.run(["op", "--version"], check=True, capture_output=True)
        except Exception as e:
            raise RuntimeError("1Password CLI not found in PATH") from e

    def _op(self, command: str, *args: str) -> str:
        proc = subprocess.run(
            ["op", command, *args],
            capture_output=True,
            text=True,
            check=False,
        )

        if proc.returncode != 0:
            raise RuntimeError(f"Error running op command: {proc.stderr}")

        return proc.stdout

    @override
    @classmethod
    def validate_key(cls, key: dict[str, Any]) -> OnePasswordProviderItemIdentifier:
        return OnePasswordProviderItemIdentifier.model_validate(key)

    @override
    async def get_file(self, item_identifier: ItemKey) -> bytes:
        parsed_key = parse_item_key(item_identifier, OnePasswordProviderItemIdentifier)
        try:
            return self._op("read", parsed_key.reference_uri).encode()

        except Exception as e:
            raise RuntimeError(f"Secret {parsed_key.reference_uri} couldn't be fetched") from e



import questionary

from terraflex.cli.builders.encryption_transformer import add_encryption_transformer
from terraflex.cli.builders.storage_provider import create_storage_provider_and_key
from terraflex.server.config import (
    ConfigFile,
    StackConfig,
)
from terraflex.utils.dependency_manager import DependenciesManager


async def start_configfile_creation_wizard(manager: DependenciesManager) -> tuple[str, ConfigFile]:
    # ask for the name of the stack
    stack_name = await questionary.text(
        "What is the name of the stack?",
        default="main",
    ).ask_async()
    if not stack_name:
        raise ValueError("Invalid stack name")

    # ask how to store the state
    storage_provider_name, storage_provider, key_identifier = await create_storage_provider_and_key(
        possible_providers=["Git", "Local"],
        main_question="How do you want to store the terraform state?",
        key_question="What is the location of the state file?",
        default_key_path="terraform.tfstate",
    )

    stack_config = StackConfig(
        state_storage=key_identifier,
        transformers=[],
    )

    result_file = ConfigFile(
        version="2",
        storage_providers={
            storage_provider_name: storage_provider,
        },
        transformers={},
        stacks={
            stack_name: stack_config,
        },
    )

    # ask if the user wants to add encryption
    should_add_encryption = await questionary.confirm(
        "Do you want to add encryption to the state?",
        default=True,
    ).ask_async()
    if should_add_encryption:
        encryption_name = await add_encryption_transformer(storage_provider_name, result_file, manager)
        if encryption_name is not None:
            stack_config.transformers.append(encryption_name)

    return stack_name, result_file



from contextlib import asynccontextmanager
from pathlib import Path
from typing import Annotated, Any, AsyncIterator, Literal, Optional, TypedDict

import uvicorn
import yaml
from fastapi import Body, Depends, FastAPI, HTTPException, Query, Request, status
from fastapi import Path as PathDep
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from terraflex.server.base_state_lock_provider import (
    Data,
    LockBody,
    LockingError,
    StateLockProviderProtocol,
)
from terraflex.server.config import ConfigFile, Settings
from terraflex.server.storage_provider_base import (
    STORATE_PROVIDERS_ENTRYPOINT,
    StorageProviderProtocol,
    WriteableStorageProviderProtocol,
)
from terraflex.server.tf_state_lock_controller import (
    TFStack,
    TFStateLockController,
)
from terraflex.server.transformation_base import (
    TRANSFORMERS_ENTRYPOINT,
    TransformerProtocol,
)
from terraflex.utils.dependency_downloader import DependencyDownloader
from terraflex.utils.dependency_manager import DependenciesManager
from terraflex.utils.plugins import get_providers, get_providers_instances

config = Settings()  # type: ignore


DEPENDENCIES_ENTRYPOINT = "terraflex.plugins.dependencies"

CONFIG_FILE_NAME = "terraflex.yaml"


async def initialize_manager() -> DependenciesManager:
    dependencies_providers = get_providers_instances(
        DependencyDownloader,
        DEPENDENCIES_ENTRYPOINT,
    )

    manager = DependenciesManager(
        dependencies=[downloader.instance for downloader in dependencies_providers.values()],
        dest_folder=config.state_dir,
    )
    await manager.initialize()
    return manager


async def generate_transformers(
    config: ConfigFile,
    manager: DependenciesManager,
    storage_providers: dict[str, StorageProviderProtocol],
    workdir: Path,
) -> dict[str, TransformerProtocol]:
    transformer_providers = get_providers(
        TransformerProtocol,
        TRANSFORMERS_ENTRYPOINT,
    )

    transformers: dict[str, TransformerProtocol] = {}
    for name, transformer in config.transformers.items():
        if transformer.type not in transformer_providers:
            raise ValueError(f"Unsupported transformer type: {transformer.type}")

        transformer_class = transformer_providers[transformer.type].model_class
        transformers[name] = await transformer_class.from_config(
            transformer.model_extra or {},
            storage_providers=storage_providers,
            manager=manager,
            workdir=workdir,
        )

    return transformers


async def create_storage_providers(
    config: ConfigFile,
    manager: DependenciesManager,
    workdir: Path,
) -> dict[str, StorageProviderProtocol]:
    storage_providers = get_providers(
        StorageProviderProtocol,
        STORATE_PROVIDERS_ENTRYPOINT,
    )

    result_storage_providers: dict[str, StorageProviderProtocol] = {}
    for name, storage_config in config.storage_providers.items():
        if storage_config.type not in storage_providers:
            raise ValueError(f"Unsupported storage provider type: {storage_config.type}")

        storage_class = storage_providers[storage_config.type].model_class
        result_storage_providers[name] = await storage_class.from_config(
            storage_config.model_extra or {},
            manager=manager,
            workdir=workdir,
        )

    return result_storage_providers


async def generate_stacks(
    config: ConfigFile,
    storage_providers: dict[str, StorageProviderProtocol],
    transformers: dict[str, TransformerProtocol],
) -> dict[str, TFStack]:
    result: dict[str, TFStack] = {}

    for stack_name, stack in config.stacks.items():
        state_storage_provider = storage_providers.get(stack.state_storage.provider)
        if state_storage_provider is None:
            raise ValueError(f"Undeclared storage provider: {stack.state_storage.provider}")

        stack_transformers: list[TransformerProtocol] = []
        for transformer_name in stack.transformers:
            transformer = transformers.get(transformer_name)
            if transformer is None:
                raise ValueError(f"Undeclared transformer: {transformer_name}")

            stack_transformers.append(transformer)

        state_key = state_storage_provider.validate_key(stack.state_storage.params or {})
        if not isinstance(state_storage_provider, WriteableStorageProviderProtocol):
            raise ValueError(
                f"Storage provider {stack.state_storage.provider} does not support writing - and it's required for state management"
            )

        result[stack_name] = TFStack(
            name=stack_name,
            storage_driver=state_storage_provider,
            data_transformers=stack_transformers,
            state_file_storage_identifier=state_key,
        )

    return result


async def initialize_controller() -> StateLockProviderProtocol:
    config_file_location = Path.cwd() / CONFIG_FILE_NAME
    if not config_file_location.exists():
        raise FileNotFoundError(f"Config file not found: {config_file_location}")

    content = config_file_location.read_bytes()
    obj = yaml.safe_load(content)
    file_config = ConfigFile.model_validate(obj)

    manager = await initialize_manager()

    storage_providers = await create_storage_providers(file_config, manager, workdir=config.state_dir)
    transformers = await generate_transformers(file_config, manager, storage_providers, workdir=config.state_dir)
    stacks = await generate_stacks(file_config, storage_providers, transformers)

    return TFStateLockController(stacks=stacks)


class AppState(TypedDict):
    controller: Optional[StateLockProviderProtocol]


state: AppState = {
    "controller": None,
}


@asynccontextmanager
async def lifespan(_: FastAPI) -> AsyncIterator[None]:
    state["controller"] = await initialize_controller()
    yield


def get_controller() -> StateLockProviderProtocol:
    if state["controller"] is None:
        raise ValueError("Controller not initialized")

    return state["controller"]


ControllerDependency = Annotated[StateLockProviderProtocol, Depends(get_controller)]

app = FastAPI(lifespan=lifespan)


@app.exception_handler(LockingError)
async def validation_exception_handler(_: Request, exc: LockingError) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_409_CONFLICT,
        content=jsonable_encoder({"detail": str(exc), "ID": exc.lock_id}),
    )


@app.get("/{stack_name}/state")
async def get_state(stack_name: str, controller: ControllerDependency) -> Data:
    # read the state file
    existing_state = await controller.get(stack_name)
    if existing_state is None:
        raise HTTPException(status_code=404, detail="State not found")

    return existing_state


@app.post("/{stack_name}/state")
async def update_state(
    stack_name: str,
    lock_id: Annotated[str, Query(..., alias="ID", description="ID of the state to update")],
    new_state: Annotated[Any, Body(..., description="New state")],
    controller: ControllerDependency,
) -> None:
    return await controller.put(stack_name, lock_id, new_state)


@app.delete("/{stack_name}/state")
async def delete_state(stack_name: str, controller: ControllerDependency) -> None:
    # TODO: should check if current user is holding the lock?
    lock = await controller.read_lock(stack_name)
    if lock is None:
        raise HTTPException(status_code=404, detail="State not found")

    return await controller.delete(stack_name, lock.ID)


@app.put("/{stack_name}/lock")
async def lock_state(stack_name: Annotated[str, PathDep()], body: LockBody, controller: ControllerDependency) -> None:
    return await controller.lock(stack_name, body)


@app.delete("/{stack_name}/lock")
async def unlock_state(stack_name: str, controller: ControllerDependency) -> None:
    return await controller.unlock(stack_name)


@app.get("/ready")
def ready() -> Literal["Ready"]:
    return "Ready"


def start_server(port: int) -> None:
    uvicorn.run(app, port=port)


if __name__ == "__main__":
    start_server(port=8600)



import pathlib
from typing import Optional

import questionary

from terraflex.plugins.local_storage_provider.local_storage_provider import (
    LocalStorageProviderInitConfig,
    LocalStorageProviderItemIdentifier,
)
from terraflex.server.config import (
    StorageProviderConfig,
    StorageProviderUsageConfig,
)


async def build_local_storage_provider(local_storage_default_path: Optional[str] = None) -> StorageProviderConfig:
    folder = pathlib.Path(
        await questionary.path(
            "Where is the folder located?",
            default=local_storage_default_path or "",
        ).ask_async()
    )

    return StorageProviderConfig(
        type="local",
        **LocalStorageProviderInitConfig(folder=folder).model_dump(),
    )


async def build_local_key_identifier(
    provider_name: str, path: str, question: str = "What is the location of the file?"
) -> StorageProviderUsageConfig:
    path = await questionary.text(
        question,
        default=path,
    ).ask_async()

    return StorageProviderUsageConfig(
        provider=provider_name,
        params=LocalStorageProviderItemIdentifier(path=path).model_dump(),
    )



import os
import pathlib
from typing import Any, Self, override

from pydantic import BaseModel
from terraflex.server.storage_provider_base import (
    ItemKey,
    StorageProviderProtocol,
    parse_item_key,
)
from terraflex.utils.dependency_manager import DependenciesManager


class EnvVarStorageProviderItemIdentifier(ItemKey):
    """Params required to reference an item in EnvVar storage provider.

    Attributes:
        key: The name of the environment variable to read.
    """

    key: str

    @override
    def as_string(self) -> str:
        return self.key


class EnvVarStorageProviderInitConfig(BaseModel):
    """Initialization params required to initialize EnvVar storage provider.

    EnvVar storage provider currently have no initialization params required.
    """


class EnvVarStorageProvider(StorageProviderProtocol):
    @override
    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        manager: DependenciesManager,
        workdir: pathlib.Path,
    ) -> Self:
        result = EnvVarStorageProviderInitConfig.model_validate(raw_config)
        return cls(
            **result.model_dump(),
        )

    @override
    @classmethod
    def validate_key(cls, key: dict[str, Any]) -> EnvVarStorageProviderItemIdentifier:
        return EnvVarStorageProviderItemIdentifier.model_validate(key)

    @override
    async def get_file(self, item_identifier: ItemKey) -> bytes:
        parsed_key = parse_item_key(item_identifier, EnvVarStorageProviderItemIdentifier)
        if parsed_key.key not in os.environ:
            raise FileNotFoundError(f"Environment variable {parsed_key.key} not found")

        return os.environ[parsed_key.key].encode()



import pathlib
from typing import Collection

from terraflex.utils.dependency_downloader import DependencyDownloader


class DependenciesManager:
    def __init__(
        self,
        dependencies: Collection[DependencyDownloader],
        *,
        dest_folder: pathlib.Path,
    ) -> None:
        self.dependencies = dependencies
        self.dest_folder = dest_folder

        self._resolved_dependencies: dict[str, pathlib.Path] = {}
        self._is_initialized = False

    async def initialize(self) -> None:
        for downloader in self.dependencies:
            results = await downloader.ensure_installed(self.dest_folder)

            for name, location in results.items():
                self._resolved_dependencies[name] = location

        self._is_initialized = True

    def require_dependency(self, name: str) -> pathlib.Path:
        if not self._is_initialized:
            raise RuntimeError("Dependencies have not been initialized")

        if name not in self._resolved_dependencies:
            raise ValueError(f"Dependency {name} has not been resolved")

        return self._resolved_dependencies[name]



import pathlib
from typing import Any, Self, override

from pydantic import BaseModel
from terraflex.server.base_state_lock_provider import LockBody
from terraflex.server.storage_provider_base import (
    ItemKey,
    LockableStorageProviderProtocol,
    parse_item_key,
)
from terraflex.utils.dependency_manager import DependenciesManager


class LocalStorageProviderItemIdentifier(ItemKey):
    """Params required to reference an item in Local storage provider.

    Attributes:
        path: The path to a specific file relative to the directory root,
            folders are also allowed as part of the path.
    """

    path: str

    @override
    def as_string(self) -> str:
        return self.path


class LocalStorageProviderInitConfig(BaseModel):
    """Initialization params required to initialize Local storage provider.

    Attributes:
        folder: The path to the directory where the files will be stored.
        folder_mode: The mode to set on the folder. Default: 0o700.
        file_mode: The mode to set on the files. Default: 0o600.
    """

    folder: pathlib.Path
    folder_mode: int = 0o700
    file_mode: int = 0o600


class LocalStorageProvider(LockableStorageProviderProtocol):
    def __init__(self, folder: pathlib.Path, folder_mode: int, file_mode: int) -> None:
        self.folder = folder.expanduser()
        self.folder_mode = folder_mode
        self.file_mode = file_mode

        if not self.folder.exists():
            self.folder.mkdir(parents=True, exist_ok=True)
            self.folder.chmod(self.folder_mode)

    @override
    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        manager: DependenciesManager,
        workdir: pathlib.Path,
    ) -> Self:
        result = LocalStorageProviderInitConfig.model_validate(raw_config)
        return cls(
            **result.model_dump(),
        )

    @override
    @classmethod
    def validate_key(cls, key: dict[str, Any]) -> LocalStorageProviderItemIdentifier:
        return LocalStorageProviderItemIdentifier.model_validate(key)

    @override
    async def get_file(self, item_identifier: ItemKey) -> bytes:
        parsed_key = parse_item_key(item_identifier, LocalStorageProviderItemIdentifier)
        file_name = parsed_key.path
        # read state
        state_file = self.folder / file_name
        try:
            return state_file.read_bytes()

        except FileNotFoundError as exc:
            raise FileNotFoundError(f"File {state_file} not found") from exc

    @override
    async def put_file(self, item_identifier: ItemKey, data: bytes) -> None:
        parsed_key = parse_item_key(item_identifier, LocalStorageProviderItemIdentifier)
        file_name = parsed_key.path
        # save state
        state_file = self.folder / file_name
        state_file.parent.mkdir(parents=True, exist_ok=True)
        state_file.write_bytes(data)
        state_file.chmod(self.file_mode)

    @override
    async def delete_file(self, item_identifier: ItemKey) -> None:
        parsed_key = parse_item_key(item_identifier, LocalStorageProviderItemIdentifier)
        file_name = parsed_key.path
        # delete state
        state_file = self.folder / file_name
        state_file.unlink()

    @override
    async def read_lock(self, item_identifier: ItemKey) -> LockBody:
        parsed_key = parse_item_key(item_identifier, LocalStorageProviderItemIdentifier)
        file_name = parsed_key.path
        # read lock data
        lock_file = self.folder / "locks" / f"{file_name}.lock"
        if not lock_file.exists():
            raise FileNotFoundError(f"Lock file {lock_file} not found")

        return LockBody.model_validate_json(lock_file.read_bytes())

    @override
    async def acquire_lock(self, item_identifier: ItemKey, data: LockBody) -> None:
        parsed_key = parse_item_key(item_identifier, LocalStorageProviderItemIdentifier)
        file_name = parsed_key.path
        # make sure lock folder exists
        locks_dir = self.folder / "locks"
        # write lock file
        lock_file = locks_dir / f"{file_name}.lock"
        lock_file.parent.mkdir(exist_ok=True, parents=True)
        lock_file.write_bytes(data.model_dump_json().encode())

    @override
    async def release_lock(self, item_identifier: ItemKey) -> None:
        parsed_key = parse_item_key(item_identifier, LocalStorageProviderItemIdentifier)
        file_name = parsed_key.path
        locks_dir = self.folder / "locks"
        lock_file = locks_dir / f"{file_name}.lock"
        lock_file.unlink()



from typing import Optional

import questionary

from terraflex.cli.builders.git_storage import build_git_key_identifier, build_git_storage_provider
from terraflex.cli.builders.local_storage import build_local_key_identifier, build_local_storage_provider
from terraflex.server.config import (
    StorageProviderConfig,
    StorageProviderUsageConfig,
)


async def create_storage_provider_and_key(
    possible_providers: list[str],
    main_question: str,
    key_question: str,
    default_key_path: str,
    storage_provider_name: Optional[str] = None,
    local_storage_default_path: Optional[str] = None,
) -> tuple[str, StorageProviderConfig, StorageProviderUsageConfig]:
    # ask how to store the state
    answer = await questionary.select(
        main_question,
        choices=possible_providers,
    ).ask_async()

    match answer:
        case "Git":
            # ask for the location of the git repository
            if storage_provider_name is None:
                storage_provider_name = "git"

            storage_provider = await build_git_storage_provider()
            key_identifier = await build_git_key_identifier(
                storage_provider_name,
                path=default_key_path,
                question=key_question,
            )

        case "Local":
            if storage_provider_name is None:
                storage_provider_name = "local"

            storage_provider = await build_local_storage_provider(local_storage_default_path)
            key_identifier = await build_local_key_identifier(
                storage_provider_name,
                path=default_key_path,
                question=key_question,
            )

        case _:
            raise ValueError("Invalid selection")

    return storage_provider_name, storage_provider, key_identifier



import pytest

from terraflex.plugins.encryption_transformation.encryption_transformation_provider import EncryptionTransformation


@pytest.mark.anyio
async def test_app(age_controller):
    transformation = EncryptionTransformation(
        encryption_provider=age_controller,
    )

    to_encrypt = b"hello world"
    result = await transformation.transform_write_file_content("test", to_encrypt)
    decrypted = await transformation.transform_read_file_content("test", result)
    assert decrypted == to_encrypt



from dataclasses import dataclass
from importlib.metadata import entry_points
from typing import Generic, Type, TypeVar
from venv import logger

T = TypeVar("T")


@dataclass
class ProviderInstance(Generic[T]):
    name: str
    instance: T


@dataclass
class Provider(Generic[T]):
    name: str
    model_class: Type[T]


def get_providers_instances(provider_type: Type[T], entrypoint_group: str) -> dict[str, ProviderInstance[T]]:
    raw_providers = entry_points(group=entrypoint_group)

    providers_instances = {provider.name: provider.load() for provider in raw_providers}
    all_providers: dict[str, ProviderInstance[T]] = {}

    for provider_name, provider_instance in providers_instances.items():
        # check that provider_class is subclass of provider
        if not isinstance(provider_instance, provider_type):
            logger.warning(f"Provider {provider_name} is not an instance of {provider_type}")
            continue

        all_providers[provider_name] = ProviderInstance(provider_name, provider_instance)

    return all_providers


def get_providers(provider_type: Type[T], entrypoint_group: str) -> dict[str, Provider[T]]:
    raw_providers = entry_points(group=entrypoint_group)

    providers_classes = {provider.name: provider.load() for provider in raw_providers}
    all_providers: dict[str, Provider[T]] = {}

    for provider_name, provider_class in providers_classes.items():
        # check that provider_class is subclass of provider
        if not issubclass(provider_class, provider_type):
            logger.warning(f"Provider {provider_name} is not a subclass of {provider_type}")
            continue

        all_providers[provider_name] = Provider(provider_name, provider_class)

    return all_providers



from typing import Any, Protocol, TypeAlias

from pydantic import BaseModel, ConfigDict


class LockBody(BaseModel):
    """Data struct that contains the lock information.

    This is the same data struct that is required by terraform.
    It follows the same fields names as the terraform lock file.

    See offical [source](https://github.com/hashicorp/terraform/blob/aea5c0cc180e0e6915454b3bf61f471c230c111b/internal/states/statemgr/locker.go#L129).

    Attributes:
        ID: The ID of the lock.
        Operation: The operation that is being performed.
        Who: The entity that is performing the operation.
        Version: The version of the lock.
        Created: The time when the lock was created.
    """

    model_config = ConfigDict(from_attributes=True)

    ID: str
    Operation: str
    Who: str
    Version: str
    Created: str


Data: TypeAlias = dict[str, Any]


class LockingError(Exception):
    def __init__(self, msg: str, lock_id: str) -> None:
        super().__init__(msg)
        self.lock_id = lock_id


class StateLockProviderProtocol(Protocol):
    async def get(self, stack_name: str) -> Data | None: ...
    async def put(self, stack_name: str, lock_id: str, value: Data) -> None: ...
    async def delete(self, stack_name: str, lock_id: str) -> None: ...
    async def read_lock(self, stack_name: str) -> LockBody | None: ...
    async def lock(self, stack_name: str, data: LockBody) -> None: ...
    async def unlock(self, stack_name: str) -> None: ...



import pathlib
from typing import Awaitable, Callable, Protocol
from venv import logger


def mv_executable_to_dest(src: pathlib.Path, dest: pathlib.Path) -> None:
    # check if equal to dest
    if src == dest:
        return

    # check if dest exists - remove it
    if dest.exists():
        dest.unlink()

    src.rename(dest)
    dest.chmod(0o755)


def should_download(expected_locations: dict[str, pathlib.Path]) -> bool:
    return not all(location.exists() for location in expected_locations.values())


def write_executable_to_file(output_bin: pathlib.Path, content: bytes) -> None:
    with open(output_bin, "wb") as f:
        f.write(content)

    output_bin.chmod(0o755)


class DownloaderProtocol(Protocol):
    async def __call__(self, version: str, expected_paths: dict[str, pathlib.Path]) -> None: ...


class DependencyDownloader:
    def __init__(
        self,
        names: list[str],
        version: str,
        downloader: Callable[[str, dict[str, pathlib.Path]], Awaitable[None]],
    ):
        self.names = names
        self.version = version
        self.downloader = downloader

    def get_bin_names(self) -> dict[str, str]:
        return {name: f"{name}-v{self.version}" for name in self.names}

    def get_expected_locations(self, dest_folder: pathlib.Path) -> dict[str, pathlib.Path]:
        bin_names = self.get_bin_names()
        return {name: (dest_folder / bin_name) for name, bin_name in bin_names.items()}

    async def ensure_installed(self, dest_folder: pathlib.Path) -> dict[str, pathlib.Path]:
        expected_locations = self.get_expected_locations(dest_folder)
        if should_download(expected_locations):
            # create directory if it doesn't exist
            dest_folder.mkdir(parents=True, exist_ok=True)
            return await self._download(expected_locations)

        return expected_locations

    async def _download(self, expected_locations: dict[str, pathlib.Path]) -> dict[str, pathlib.Path]:
        logger.info(f"Downloading {self.names} client...")
        await self.downloader(self.version, expected_locations)
        return expected_locations



import questionary

from terraflex.cli.builders.git_storage import build_git_key_identifier
from terraflex.cli.builders.local_storage import build_local_key_identifier
from terraflex.cli.builders.storage_provider import create_storage_provider_and_key
from terraflex.plugins.encryption_transformation.age.controller import AgeKeygenController
from terraflex.plugins.encryption_transformation.age.provider import AgeKeyConfig
from terraflex.plugins.encryption_transformation.encryption_transformation_provider import EncryptionTransformerConfig
from terraflex.server.app import (
    config as server_config,
)
from terraflex.server.app import (
    create_storage_providers,
)
from terraflex.server.config import (
    ConfigFile,
    StorageProviderUsageConfig,
    TransformerConfig,
)
from terraflex.server.storage_provider_base import WriteableStorageProviderProtocol
from terraflex.utils.dependency_manager import DependenciesManager


async def generate_encryption_key(
    manager: DependenciesManager,
    config_file: ConfigFile,
    provider_name: str,
    key_identifier: StorageProviderUsageConfig,
) -> None:
    controller_location = manager.require_dependency("age-keygen")
    controller = AgeKeygenController(
        binary_location=controller_location,
    )
    private_key = await controller.generate_key_bytes()

    storage_providers = await create_storage_providers(
        config_file,
        manager,
        workdir=server_config.state_dir,
    )

    storage_provider_instance = storage_providers[provider_name]
    item_key = storage_provider_instance.validate_key(key_identifier.params or {})
    try:
        await storage_provider_instance.get_file(item_key)
        # file exists, ask if we should replace it
        should_replace = await questionary.confirm(
            "The key already exists, do you want to replace it?",
            default=False,
        ).ask_async()
        if not should_replace:
            print("Keeping existing key...")
            return

    except FileNotFoundError:
        pass

    if isinstance(storage_provider_instance, WriteableStorageProviderProtocol):
        await storage_provider_instance.put_file(item_key, private_key)


async def add_encryption_transformer(
    state_storage_provider_name: str, config_file: ConfigFile, manager: DependenciesManager
) -> str | None:
    transformer_name = await questionary.text(
        "What is the name of the transformer?",
        default="encryption",
    ).ask_async()

    key_type = await questionary.select(
        "What type of key do you want to use for encryption?",
        choices=[
            "age",
        ],
    ).ask_async()

    match key_type:
        case "age":
            choices = [
                *[
                    questionary.Choice(f"{key} ({value.type})", value=(key, value))
                    for key, value in config_file.storage_providers.items()
                    if key != state_storage_provider_name
                ],
                questionary.Choice("Create a new storage", value="create"),
            ]

            selected_storage_provider = await questionary.select(
                "Where should the key be stored?",
                choices=choices,
            ).ask_async()

            if selected_storage_provider == "create":
                provider_name, storage_provider, key_identifier = await create_storage_provider_and_key(
                    possible_providers=["Local"],
                    main_question="How do you want to store the secret?",
                    key_question="What is the location of the secret key?",
                    default_key_path="age-key.txt",
                    storage_provider_name="encryption",
                    local_storage_default_path="~/secrets/",
                )

                config_file.storage_providers[provider_name] = storage_provider

            else:
                provider_name, storage_provider_config = selected_storage_provider
                # get a new key for the storage provider
                match storage_provider_config.type:
                    case "local":
                        key_identifier = await build_local_key_identifier(provider_name, path="age-key.txt")

                    case "git":
                        key_identifier = await build_git_key_identifier(provider_name, path="age-key.txt")

                    case _:
                        raise ValueError("Invalid selection")

            should_generate_key = await questionary.confirm(
                "Do you want to generate a new key?",
                default=True,
            ).ask_async()
            if should_generate_key:
                await generate_encryption_key(manager, config_file, provider_name, key_identifier)

            encryption_config = EncryptionTransformerConfig(
                key_type="age",
                **AgeKeyConfig(
                    import_from_storage=key_identifier,
                ).model_dump(),
            )

        case _:
            raise ValueError("Invalid selection")

    transformer_config = TransformerConfig(
        type="encryption",
        **encryption_config.model_dump(),
    )

    config_file.transformers[transformer_name] = transformer_config

    return transformer_name



import pathlib
from typing import (
    Annotated,
    Any,
    Optional,
)

import semver
import xdg_base_dirs
from pydantic_settings import (
    BaseSettings,
)

from pydantic import BaseModel, ConfigDict, Field, field_validator

PACKAGE_NAME = "terraflex"

CONFIG_VERSION = "2"


class StorageProviderUsageConfig(BaseModel):
    """Data struct that contains the parameters that link to a specific file in a given storage provider.

    Each storage provider defines it's own unique usage parameters.

    Please refer to your storage provider usage config for key specification.

    Attributes:
        provider: storage provider name defined in `storage_providers` section of the config file.
        params: storage provider specific usage parameters - each storage provider defines it's own params -
            and this dict will be processed and validated dynamically by each storage provider.

    Example:
        This is an example of a `local` storage provider usage config:
        ```yaml
        provider: local
        params:
            path: ./path/to/item.txt
        ```
    """

    provider: str
    params: Optional[dict[str, Any]] = None


class StorageProviderConfig(BaseModel):
    """Data struct that contains the configuration for a storage provider.

    Each storage provider defines it's own unique configuration parameters -
    and the parameters will be passed through to the storage provider.

    Attributes:
        type: storage provider type as declared in the entrypoint.
        **kwargs: storage provider specific configuration parameters.

    Example:
        In this example, the `local` storage provider has a `folder` parameter that is required.
        The storage provider will get a dict: `{"folder": "/path/to/folder"}` as the configuration.

        ```yaml
        type: local
        folder: /path/to/folder
        ```
    """

    model_config = ConfigDict(extra="allow")
    type: str


class TransformerConfig(BaseModel):
    """Base Transformer configuration.

    Each transformer defines it's own unique configuration parameters -
    and the parameters will be passed through to the transformer.

    Attributes:
        type: The type of the transformer.
        **kwargs: Additional configuration for the transformer.
    """

    model_config = ConfigDict(extra="allow")
    type: str


class StackConfig(BaseModel):
    """Configuration for a terraform stack.

    Attributes:
        state_storage: The storage provider configuration for the state file.
        transformers: The list of transformers to apply to the data.
    """

    state_storage: StorageProviderUsageConfig
    transformers: list[str]


class ConfigFile(BaseModel):
    """The configuration file for terraflex.

    Attributes:
        version: The version of the configuration file.
        storage_providers: The configuration for the storage providers.
        transformers: The configuration for the transformers.
        stacks: The configuration for the stacks.
    """

    version: str = CONFIG_VERSION
    storage_providers: dict[str, StorageProviderConfig]
    transformers: dict[str, TransformerConfig]
    stacks: dict[str, StackConfig]

    @field_validator("version")
    @classmethod
    def validate_version(cls, value: str) -> str:
        current_version = semver.Version.parse(value, optional_minor_and_patch=True)
        config_version = semver.Version.parse(CONFIG_VERSION, optional_minor_and_patch=True)
        if current_version < config_version:
            raise ValueError(
                f"Unsupported version ({current_version} < {config_version}) - please upgrade the config file"
            )

        if current_version > config_version:
            raise ValueError(
                f"Unsupported version ({current_version} > {config_version}) - please check if there is a newer version of {PACKAGE_NAME}"
            )

        return value


class Settings(BaseSettings):
    state_dir: Annotated[
        pathlib.Path,
        Field(
            default=xdg_base_dirs.xdg_data_home() / PACKAGE_NAME,
        ),
    ]



import pathlib
from typing import Any, Self, override

from pydantic import BaseModel, ConfigDict
from terraflex.plugins.encryption_transformation.encryption_base import EncryptionProtocol
from terraflex.server.storage_provider_base import StorageProviderProtocol
from terraflex.server.transformation_base import (
    TransformerProtocol,
)
from terraflex.utils.dependency_manager import DependenciesManager
from terraflex.utils.plugins import get_providers

ENCRYPTION_PROVIDER_ENTRYPOINT = "terraflex.plugins.transformer.encryption"


class EncryptionTransformerConfig(BaseModel):
    """Transformer that encrypts and decrypts the content of the files using the specified encryption provider.

    Attributes:
        key_type: The type of the encryption key.
        **kwargs: Additional configuration for the encryption provider.

    Example:
        Encryption transformer with `age` encryption provider:
        ```yaml
        type: encryption
        key_type: age
        import_from_storage:
            provider: envvar
            params:
                key: AGE_PRIVATE_KEY
        ```
    """

    model_config = ConfigDict(extra="allow")
    key_type: str


encryption_providers = get_providers(EncryptionProtocol, ENCRYPTION_PROVIDER_ENTRYPOINT)


class EncryptionTransformation(TransformerProtocol):
    def __init__(self, encryption_provider: EncryptionProtocol):
        self.encryption_provider = encryption_provider

    @override
    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        storage_providers: dict[str, StorageProviderProtocol],
        manager: DependenciesManager,
        workdir: pathlib.Path,
    ) -> Self:
        config = EncryptionTransformerConfig.model_validate(raw_config)
        encryption_provider = encryption_providers.get(config.key_type)
        if encryption_provider is None:
            raise ValueError(f"Unsupported encryption key type: {config.key_type}")

        encryption_controller_class = encryption_provider.model_class
        controller = await encryption_controller_class.from_config(
            raw_config,
            storage_providers=storage_providers,
            manager=manager,
        )

        return cls(encryption_provider=controller)

    @override
    async def transform_read_file_content(self, file_identifier: str, content: bytes) -> bytes:
        return await self.encryption_provider.decrypt(file_identifier, content)

    @override
    async def transform_write_file_content(self, file_identifier: str, content: bytes) -> bytes:
        return await self.encryption_provider.encrypt(file_identifier, content)



from typing import Any, Self, override

from pydantic import BaseModel
from terraflex.plugins.encryption_transformation.age.controller import AgeController, AgeKeygenController
from terraflex.plugins.encryption_transformation.age.downloader import AgeDownloader
from terraflex.plugins.encryption_transformation.encryption_base import EncryptionProtocol
from terraflex.server.config import StorageProviderUsageConfig
from terraflex.server.storage_provider_base import StorageProviderProtocol
from terraflex.utils.dependency_downloader import DependencyDownloader
from terraflex.utils.dependency_manager import DependenciesManager


class AgeKeyConfig(BaseModel):
    """Configuration for the Age encryption provider.

    Attributes:
        import_from_storage: usage reference to the storage provider where the private key is stored.
    """

    import_from_storage: StorageProviderUsageConfig


AgeDependency = DependencyDownloader(
    names=["age", "age-keygen"],
    version="1.2.0",
    downloader=AgeDownloader(),
)


class AgeEncryptionProvider(EncryptionProtocol):
    def __init__(
        self,
        controller: AgeController,
    ):
        self.controller = controller

    @classmethod
    async def from_config(
        cls,
        raw_config: Any,
        *,
        storage_providers: dict[str, StorageProviderProtocol],
        manager: DependenciesManager,
    ) -> Self:
        config = AgeKeyConfig.model_validate(raw_config)
        storage_provider = storage_providers.get(config.import_from_storage.provider)
        if storage_provider is None:
            raise ValueError(f"Undeclared storage provider: {config.import_from_storage.provider}")

        storage_params = config.import_from_storage.params
        if storage_params is None:
            raise ValueError("Missing storage params")

        storage_key = storage_provider.validate_key(storage_params)
        private_key = await storage_provider.get_file(storage_key)
        age_keygen_controller = AgeKeygenController(
            binary_location=manager.require_dependency("age-keygen"),
        )
        public_key = await age_keygen_controller.get_public_key_from_bytes(private_key)

        return cls(
            controller=AgeController(
                binary_location=manager.require_dependency("age"),
                private_key=private_key,
                public_key=public_key,
            )
        )

    @override
    async def encrypt(self, file_name: str, content: bytes) -> bytes:
        return await self.controller.encrypt(file_name, content)

    @override
    async def decrypt(self, file_name: str, content: bytes) -> bytes:
        return await self.controller.decrypt(file_name, content)
